#! /usr/bin/env luajit


-----------------------------------------------------------------
--Original code by MC-CNN
-- Edited by Storm Westlie











-----------------------------------------------------------------



require 'torch'

io.stdout:setvbuf('no')
for i = 1,#arg do
   io.write(arg[i] .. ' ')
end
io.write('\n')
dataset = table.remove(arg, 1)
arch = table.remove(arg, 1)
assert(dataset=='modelfish_test')
assert(arch == 'fast')

cmd = torch.CmdLine()
cmd:option('-gpu', 1, 'gpu id')
cmd:option('-seed', 42, 'random seed')
cmd:option('-debug', false)
cmd:option('-d', 'kitti | mb')
cmd:option('-a', 'train_tr | train_all | test_te | test_all | submit | time | predict', 'train')
cmd:option('-net_fname', '')
cmd:option('-make_cache', false)
cmd:option('-use_cache', false)
cmd:option('-print_args', false)
cmd:option('-sm_terminate', '', 'terminate the stereo method after this step')
cmd:option('-sm_skip', '', 'which part of the stereo method to skip')
cmd:option('-tiny', false)
cmd:option('-subset', 1)
cmd:option('-left', '')
cmd:option('-right', '')
cmd:option('-disp_max', '')
cmd:option('-rect', 'imperfect')
cmd:option('-color', 'gray')
cmd:option('-epochs', 50, 'Number of epochs to train for')
cmd:option('-epoch_step', 15, '1/2 learningrate at every epoch_step')


opt = cmd:parse(arg)


require 'cunn'
require 'cutorch'
require 'image'
require 'libadcensus'
require 'libcv'
require 'cudnn'
require 'optim'
cudnn.benchmark = true

include('Margin2.lua')
include('Normalize2.lua')
include('StereoJoin.lua')
include('StereoJoin1.lua')
include('SpatialConvolution1_fw.lua')
-- include('SpatialLogSoftMax.lua')

torch.manualSeed(opt.seed)
cutorch.manualSeed(opt.seed)
cutorch.setDevice(tonumber(opt.gpu))



cmd_str = dataset .. '_' .. arch
for i = 1,#arg do
   cmd_str = cmd_str .. '_' .. arg[i]
end

function isnan(n)
   return tostring(n) == tostring(0/0)
end

function fromfile(fname)
   local file = io.open(fname .. '.dim')
   local dim = {}
   for line in file:lines() do
      table.insert(dim, tonumber(line))
   end
   if #dim == 1 and dim[1] == 0 then
      return torch.Tensor()
   end

   local file = io.open(fname .. '.type')
   local type = file:read('*all')

   local x
   if type == 'float32' then
      x = torch.FloatTensor(torch.FloatStorage(fname))
   elseif type == 'int32' then
      x = torch.IntTensor(torch.IntStorage(fname))
   elseif type == 'int64' then
      x = torch.LongTensor(torch.LongStorage(fname))
   else
      print(fname, type)
      assert(false)
   end

   x = x:reshape(torch.LongStorage(dim))
   return x
end

function get_window_size(net)
   ws = 1
   for i = 1,#net.modules do
      local module = net:get(i)
      if torch.typename(module) == 'cudnn.SpatialConvolution' then
         ws = ws + module.kW - 1
      end
   end
   return ws
end

-- load testing data

if dataset == 'modelfish_test' then
   height = 300
   width = 1330
   disp_max = 170
   n_input_plane = 1
   err_at = 3

   if opt.a == 'test' then
      if dataset == 'modelfish_test' then
         X0 = fromfile('data.modelfish2019_test/x0.bin')
         X1 = fromfile('data.modelfish2019_test/x1.bin')
         dispnoc = fromfile('data.modelfish2019_test/dispnoc.bin')
         metadata = fromfile('data.modelfish2019_test/metadata.bin')
         te = fromfile('data.modelfish2019_test/te.bin')
      end
   end
end


function print_net(net)
   local s
   local t = torch.typename(net) 
   if t == 'cudnn.SpatialConvolution' then
      print(('conv(in=%d, out=%d, k=%d)'):format(net.nInputPlane, net.nOutputPlane, net.kW))
   elseif t == 'nn.SpatialConvolutionMM_dsparse' then
      print(('conv_dsparse(in=%d, out=%d, k=%d, s=%d)'):format(net.nInputPlane, net.nOutputPlane, net.kW, net.sW))
   elseif t == 'cudnn.SpatialMaxPooling' then
      print(('max_pool(k=%d, d=%d)'):format(net.kW, net.dW))
   elseif t == 'nn.StereoJoin' then
      print(('StereoJoin(%d)'):format(net.disp_max))
   elseif t == 'nn.Margin2' then
      print(('Margin2(margin=%f, pow=%d)'):format(opt.m, opt.pow))
   elseif t == 'nn.GHCriterion' then
      print(('GHCriterion(m_pos=%f, m_neg=%f, pow=%d)'):format(opt.m_pos, opt.m_neg, opt.pow))
   elseif t == 'nn.Sequential' then
      for i = 1,#net.modules do
         print_net(net.modules[i])
      end
   else
      print(net)
   end
end

function clean_net(net)
   net.output = torch.CudaTensor()
   net.gradInput = nil
   net.weight_v = nil
   net.bias_v = nil
   net.gradWeight = nil
   net.gradBias = nil
   net.iDesc = nil
   net.oDesc = nil
   net.finput = torch.CudaTensor()
   net.fgradInput = torch.CudaTensor()
   net.tmp_in = torch.CudaTensor()
   net.tmp_out = torch.CudaTensor()
   if net.modules then
      for _, module in ipairs(net.modules) do
         clean_net(module)
      end
   end
   return net
end

function save_net(epoch)
   if arch == 'slow' then
      obj = {clean_net(net_te), clean_net(net_te2), opt}
   elseif arch == 'fast' then
      obj = {clean_net(net_te), opt}
   end
   if epoch == 0 then
      fname = ('net/net_%s.t7'):format(cmd_str)
   else
      fname = ('net/net_%s_%d.t7'):format(cmd_str, epoch)
   end
   torch.save(fname, obj, 'ascii')
   return fname
end

function forward_free(net, input)
   local currentOutput = input
   for i=1,#net.modules do
      net.modules[i].oDesc = nil
      local nextOutput = net.modules[i]:updateOutput(currentOutput)
      if currentOutput:storage() ~= nextOutput:storage() then
         currentOutput:storage():resize(1)
         currentOutput:resize(0)
      end
      currentOutput = nextOutput
   end
   net.output = currentOutput
   return currentOutput
end

function fix_border(net, vol, direction)
   local n = (get_window_size(net) - 1) / 2
   for i=1,n do
      vol[{{},{},{},direction * i}]:copy(vol[{{},{},{},direction * (n + 1)}])
   end
end

function stereo_predict(x_batch, id)
   local vols, vol

   if arch == 'fast' then
      forward_free(net_te_test, x_batch:clone())
      vols = torch.CudaTensor(2, disp_max, x_batch:size(3), x_batch:size(4)):fill(0 / 0)
      adcensus.StereoJoin(net_te_test.output[{{1}}], net_te_test.output[{{2}}], vols[{{1}}], vols[{{2}}])
      fix_border(net_te_test, vols[{{1}}], -1)
      fix_border(net_te_test, vols[{{2}}], 1)
      clean_net(net_te_test)
   end

   disp = {}
   local mb_directions = opt.a == 'predict' and {1, -1} or {-1}
   for _, direction in ipairs(dataset == 'mb' and mb_directions or {1, -1}) do

      
      if arch == 'fast' then
         vol = vols[{{direction == -1 and 1 or 2}}]
      end

      _, d = torch.min(vol, 2)
      disp[direction == 1 and 1 or 2] = d:cuda():add(-1)
   end

   collectgarbage()

   return disp[2]
end

if opt.a == 'test' then


      test_fname = opt.net_fname
      net_te_test = torch.load(test_fname, 'ascii')[1]
   	net_te_test.modules[#net_te_test.modules] = nil

    --val accuracy 
   	examples = torch.totable(te)
   	err_sum = 0
		x_batch = torch.CudaTensor(2, 1, height, width)
		pred_good = torch.CudaTensor()
		pred_bad = torch.CudaTensor()
		for _, i in ipairs(examples) do
            print(i)
		      img_height = metadata[{i,1}]
		      img_width = metadata[{i,2}]
		      id = metadata[{i,3}]
		      x0 = X0[{i,{},{1,img_height},{1,img_width}}]
		      x1 = X1[{i,{},{1,img_height},{1,img_width}}]

		   x_batch:resize(2, 1, img_height, img_width)
		   x_batch[1]:copy(x0)
		   x_batch[2]:copy(x1)

		   collectgarbage()
		   cutorch.synchronize()
		   sys.tic()
		   pred = stereo_predict(x_batch, id)
		   cutorch.synchronize()
		   runtime = sys.toc()


	      assert(not isnan(pred:sum()))

	      actual = dispnoc[{i,{},{1,img_height},{1,img_width}}]:cuda()

	      pred_good:resizeAs(actual)
	      mask = torch.CudaTensor():resizeAs(actual):ne(actual, 0)
	      actual:add(-1, pred):abs()
	      pred_bad:gt(actual, err_at):cmul(mask)
	      local err = pred_bad:sum() / mask:sum()
	      err_sum = err_sum + err
		end
		mean_test_acc = (1.0-err_sum / #examples)
		print('Test mean accuracy', mean_test_acc, 'Error thresh:', err_at)

     	collectgarbage()
   

   collectgarbage()
end




